#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# Copyright (c) 2009 Efrim <efrim@azylum.org> {{{
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public
#  License along with this program.
#  If not, see <http://www.gnu.org/licenses/>.
#
#  }}}

"""
This code is dealing with your MPD server.
It will add automagicaly track to the playlist.
Simply run:
    python mpd_sima.py --pid=/path/to/pid_file.pid
"""

from __future__ import (with_statement)

__version__ = u'0.4.0beta'
__revison__ = u'$Revision: 138 $'[11:-2]
__author__ = u'$Author: kaliko $'
__date__ = (
    u"$LastChangedDate: 2009-07-30 14:58:51 +0200 (jeu 30 jui 2009) $"[18:28])

# IMPORTS {{{
import pickle
import random
import signal
import sys
import time

from datetime import date
from mpd import ConnectionError
from os.path import (basename, isfile, getsize)
from types import ListType

from lib.audioscrobbler import (AudioScrobblerQuery,
        AudioScrobblerConnectionError)
import lib.mpdConnector as mpdConnector
from utils.config import (ConfMan, load_cache, load_usr_cache, CacheError)
from utils.leven import levenshtein_ratio
from utils.logutil import (sima_logger, LEVELS)
from utils.startopt import StartOpt

#}}}


class SimaClass():#Main class dealing with Similar artists to add to MPD {{{
    """
    This is a simple Class to feed music in mpd playlist with suggestions
    from lastfm/libre.fm.

    ---------------------------------------------------------------------
    mpd_add_history = { 'artist0': ["/file0", "file1"],
                        'artist1': ["/fileA", "fileB", "fileC"]
                      }
    ---------------------------------------------------------------------
    """

    def __init__(self, config, mpd=None):#{{{
        self.current_file = None
        self.sima_config = config
        self.mpd_connection = mpd
        self.tracks_to_add = []
        self.mpd_add_history = {}
        self.last_as_error = None#}}}

    def inithist(self, hist_file):#{{{unicodeReady
        """
        Load history.
        """
        if isfile(hist_file) and getsize(hist_file) != 0:
            with open(hist_file, 'rb') as hist_fd:
                SIMA_LOGGER.info(u'Loading history from: %s' % hist_file)
                self.mpd_add_history = pickle.load(hist_fd)
                SIMA_LOGGER.info(u'%d artists in history' %
                        len(self.mpd_add_history))
        elif not isfile(hist_file):
            SIMA_LOGGER.info(u'Will create history file in %s' % hist_file)
        #}}}

    def writehist(self, hist_file):#{{{unicodeReady
        """
        Write history to file.
        """
        with open(hist_file, 'wb') as hist_fd:
            SIMA_LOGGER.info(u'Saving history to %s' % hist_file)
            pickle.dump(self.mpd_add_history, hist_fd, -1)
            #}}}

    def addhist(self, track):#{{{unicodeReady
        """
        Deal with history.
        Try to add file to history, return False if already in
        True either.
        May be used to add AND check history at the same time.
        """
        # TODO: Bug report #685
        track = artist_nolist(track)
        mpd_file, mpd_artist = track['file'], track.get('artist', None)
        SIMA_LOGGER.debug(u'Checking history for %s' %
                          unicode(basename(mpd_file),'utf-8'))

        if mpd_artist in self.mpd_add_history and \
           mpd_file not in self.mpd_add_history[mpd_artist]:
            self.mpd_add_history[mpd_artist].append(mpd_file)
            format_list = [unicode('"'+basename(mfile)+'"','utf-8') for
                           mfile in self.mpd_add_history[mpd_artist]]
            format_list = ' / '.join(format_list)
            SIMA_LOGGER.debug(u'Not in history, appending file: %s' %
                              format_list)
            #SIMA_LOGGER.debug(u'not in history, appending file.')
            return True
        elif mpd_artist not in self.mpd_add_history:
            SIMA_LOGGER.debug(u'Track never played before.')
            self.mpd_add_history[mpd_artist] = []
            self.mpd_add_history[mpd_artist].append(mpd_file)
            #format_list = [basename(mfile) for mfile in
            #              self.mpd_add_history[mpd_artist]]
            SIMA_LOGGER.debug(u'not in history, appending file: %s' %
                              unicode(basename(mpd_file),'utf-8'))
            return True
        else:
            return False#}}}

    def is_inqueue(self, track):#{{{unicodeReady
        """
        Check if track is in the queue.
        """
        cursonpos = int(self.mpd_connection.currentsong().get('pos',0))
        playlist_dict = self.mpd_connection.playlistinfo()
        #print mpd_con.mpd_cli.playlistinfo()
        if track['file'] in [pl_track['file'] for pl_track in playlist_dict[cursonpos:]]:
            SIMA_LOGGER.debug(u'"%s" already in the queue' %
                              unicode(basename(track['file']),'utf-8'))
            return True
        return False#}}}

    def is_inhist(self, track):#{{{unicodeReady
        """
        Check if track is in history.
        """
        # TODO: Bug report #685
        track = artist_nolist(track)
        mpd_file, mpd_artist = track['file'], track['artist']
        if mpd_artist in self.mpd_add_history and \
           mpd_file in self.mpd_add_history[mpd_artist]:
            #SIMA_LOGGER.debug(u'Already in history (along with %i other files).'
            #                  % len(self.mpd_add_history[mpd_artist]))
            SIMA_LOGGER.debug(u'File already in history: %s' %
                              unicode(basename(track['file']),'utf-8'))
            return True
        SIMA_LOGGER.debug(u'\o/ File not in history: %s' %
                          unicode(basename(track['file']),'utf-8'))
        return False#}}}

    def extract_unplayed(self, tracks):#{{{
        """
        Extract one unplayed track for an artist.
        Check against history and file in queue.
        """
        selected_track = None
        for track in tracks:
            # TODO: Bug report #685
            track = artist_nolist(track)
            if not self.is_inqueue(track) and \
               track not in self.tracks_to_add:
                if not self.is_inhist(track):
                    selected_track = track
                    break
        return selected_track
    #}}}

    def maddtrk(self):#{{{unicodeReady
        """
        Add track to MPD.
        """
        tracks = self.tracks_to_add
        for track in tracks:
            if track.get('title'):
                SIMA_LOGGER.info(u'# Add to playlist: %s / %s' %
                    (unicode(track['artist'],'utf-8'),
                     unicode(track['title'],'utf-8')))
            elif not track.get('title'):
                SIMA_LOGGER.info(
                    u'##Add to playlist: %s / %s (#MISSING title tag)' %
                    (unicode(track['artist'],'utf-8'),
                     unicode(basename(track['file']),'utf-8')))
                SIMA_LOGGER.warning(u'MISSING TITLE TAG in: %s' % 
                                    unicode(track['file'],'utf-8'))
            #DEV##ADD#
            #self.mpd_connection.add(track['file'])#}}}

    def msearch(self, similar=None):#{{{
        """
        Look for similar artists in MPD library.
        """
        mpd_artists = []
        mpd_missing_artists = []
        #for as_artist in self.as_artist_names:
        # Look only for the first fifty artists #PARAM#
        for as_artist in similar[0:50]:
            #SIMA_LOGGER.debug(u'Looking for %s' % unicode(as_artist,'utf-8'))#DEV#
            mpd_search = self.mpd_connection.search('artist', as_artist)
            if not mpd_search:
                # High verbosity logger
                #SIMA_LOGGER.debug(u'Nothing found with fuzzy search')#DEV#
                continue
            SIMA_LOGGER.debug(u'Found %03i fuzzy match(es) for artist "%s"' %
                              (len(mpd_search), unicode(as_artist,'utf-8')))
            for result in mpd_search:
                # TODO: Bug report #685
                result = artist_nolist(result)
                if 'artist' in result and \
                   result['artist'] not in mpd_artists:
                    leven_ratio = levenshtein_ratio(as_artist.lower(),
                                                    result['artist'].lower())
                    if leven_ratio == 1:
                        mpd_artists.append(result['artist'])
                    elif leven_ratio >= 0.8:#PARAM#
                        SIMA_LOGGER.debug(
                            u'FZZ: "%s" should match "%s" (lr=%1.2f)' %
                            (unicode(as_artist,('utf-8')),
                             unicode(result['artist'],'utf-8'), leven_ratio))
                        mpd_artists.append(result['artist'])
                    elif result['artist'] not in mpd_missing_artists:
                        SIMA_LOGGER.debug(
                            u'FZZ: "%s" does not match "%s" (lr="%1.2f")' %
                            (unicode(result['artist'],'utf-8'),
                             unicode(as_artist,'utf-8'), leven_ratio))
                        mpd_missing_artists.append(result['artist'])
        return mpd_artists#}}}

    def needadd(self):#{{{
        """Controls either or not new tracks are needed."""
        mpd_status = self.mpd_connection.status()
        queue_lgh = self.sima_config.getint('sima', 'queue_length')
        mfile = self.mpd_connection.currentsong().get('file', None)

        if (int(mpd_status['playlistlength']) - int(mpd_status['song']) - 1 >
            int(queue_lgh)):
            return False

        # DOES or DOES NOT continue if the same title is still playing :/ !!??
        #if mfile == self.current_file:
        #    return False

        return True#}}}

    def sima_lp(self):#{{{unicodeReady
        """
        Main Loop.
        """
        nbtracks2add = self.sima_config.getint('sima', 'track_to_add')
        #mpd_track = self.mpd_connection.currentsong()
        #playlist_length = int(mpd_status['playlistlength'])
        #mpd_track_position = int(mpd_status['song']) + 1
        # Test whether or not artist/title tags are present
        title = self.mpd_connection.currentsong().get('title', None)
        artist = self.mpd_connection.currentsong().get('artist', None)
        mfile = self.mpd_connection.currentsong().get('file', None)

        # TODO: Bug report #685
        if type(artist) is ListType:
            artist = ' '.join(artist)
        u_mfile = unicode(mfile,'utf-8')

        if not artist:
            if mfile != self.current_file:
                SIMA_LOGGER.warning(u'Playing %s' % u_mfile)
                SIMA_LOGGER.warning(u'No artist tag set for %s' % u_mfile)
                SIMA_LOGGER.warning(u'Cannont look for similar artist.')
                self.current_file = mfile
                self.addhist(self.mpd_connection.currentsong())
            return False

        #Set unicode strings
        if not title:
            SIMA_LOGGER.warning(u'No title tag set for "%s".' % basename(u_mfile))
            u_title = None
        elif title:
            u_title = unicode(title,'utf-8')
        u_artist = unicode(artist,'utf-8')

        if mfile != self.current_file:
            self.current_file = mfile
            SIMA_LOGGER.info(u'Playing %s / %s (%d tracks in history for %s)' %
                             (u_artist, u_title,
                              len(self.mpd_add_history.get(artist, [])),
                              u_artist))
            self.addhist(self.mpd_connection.currentsong())

        if not self.needadd():
            return False

        # MPD state
        #self.current_title = title
        #self.current_artist = artist
        SIMA_LOGGER.debug(u'Seeking for new tracks to add.')
        # Look for similar artists
        similarts = []
        try:
            similarts = similar_artist(artist)
        except AudioScrobblerConnectionError, error:
            if self.last_as_error and (time.time() - self.last_as_error<360):
                # Time check, not to flood the log output.
                # print out a warning each 6 minutes (360 secs)
                # uncomment next line to have a single warning
                #self.last_as_error = time.time()
                return False
            SIMA_LOGGER.warning(u'Scrobbler server replies: %s…' % error)
            self.last_as_error = time.time()
            return False
        sima_str = unicode('/'.join(similarts[0:3*nbtracks2add]),'utf-8')
        SIMA_LOGGER.info(u'some similar artist(s) found: "%s/…"' %
                (sima_str))

        # search similar artists in mpd database
        SIMA_LOGGER.debug(u'Checking in MPD database…')
        mpd_artist_found = self.msearch(similarts)
        if not mpd_artist_found:
            SIMA_LOGGER.info(u'Nothing similar found in MPD for "%s".' %
                             u_artist)
            return False
        sima_str = unicode('/'.join(mpd_artist_found),'utf-8')
        SIMA_LOGGER.info(u'Found %d similar artist(s) in MPD: "%s"' %
                         (len(mpd_artist_found), sima_str))
        # padd list of artist if not enought artists are found
        while len(mpd_artist_found) < nbtracks2add:
            mpd_artist_found.extend(mpd_artist_found)
        sima_str = unicode('/'.join(mpd_artist_found),'utf-8')
        SIMA_LOGGER.debug(u'Artists list, possibly padded: %s' %
                          sima_str)
        random.shuffle(mpd_artist_found)
        self.tracks_to_add = []
        for mfd_artist in mpd_artist_found:
            mpd_find = self.mpd_connection.find('artist', mfd_artist)
            u_mfd_artist = unicode(mfd_artist,'utf-8')
            SIMA_LOGGER.debug(u'Trying to find titles to add for "%s"' %
                              u_mfd_artist)
            random.shuffle(mpd_find)
            unplayed_track = self.extract_unplayed(mpd_find)
            if not unplayed_track:
                SIMA_LOGGER.info(u'Unable to find title to add'+
                                 u' for "%s".' % u_mfd_artist)
                if mfd_artist in self.mpd_add_history:
                    SIMA_LOGGER.info(
                            u'History already gathers %i titles for %s.' %
                            (len(self.mpd_add_history[mfd_artist]),
                             u_mfd_artist))
            else:
                self.tracks_to_add.append(unplayed_track)
            if len(self.tracks_to_add) == nbtracks2add:
                break
        if not self.tracks_to_add:
            SIMA_LOGGER.warning(u'Found no unplayed tracks, your '+
                                u'history is probably getting large.')
            return False

        SIMA_LOGGER.info(u'Found %i unplayed track(s) (target %i)' %
                         (len(self.tracks_to_add), nbtracks2add))
        self.maddtrk()#}}}


#}}}


# Some FUNCTIONS #{{{


def artist_nolist(track):#{{{unicodeReady
    """
    Stupid bug for artist like "Fèqadu Amdè - Mesqel"
    Black Stripes reported the same issue on "Beth Gibbons & Rustin Man"
    result['artist'] is a list, hence no lower method.
    * 07/29/2009 Add workaround for title.
    # TODO: Bug report #685
    """
    if 'artist' not in track:
        return track
    if type(track['artist']) is ListType:
        u_artist = unicode(' * '.join(track['artist']),'utf-8')
        SIMA_LOGGER.debug(u'## damn entry to format: \t %s' % u_artist)
        track['artist'] = ' '.join(track['artist'])
    if 'title' not in track:
        return track
    if type(track['title']) is ListType:
        u_title = unicode(' * '.join(track['title']),'utf-8')
        SIMA_LOGGER.debug(u'## damn entry to format: \t %s' % u_title)
        track['title'] = ' '.join(track['title'])
    return track#}}}


def writecache(cache_file):#{{{unicodeReady
    """
    Write cache to file.
    TODO: controls whether or not cache_file encoding matters here.
    """
    if isfile(cache_file):
        with open(cache_file, 'wb') as cache_fd:
            SIMA_LOGGER.info(u'Updating cache with latest requests: %s' %
                             cache_file)
            pickle.dump(CACHE, cache_fd, -1)
            pickle.dump(CACHE_STAMP, cache_fd, -1)
    elif not isfile(cache_file):
        with open(cache_file, 'wb') as cache_fd:
            cache_stamp = {}
            cache_stamp['DATE'] = date.today()
            SIMA_LOGGER.info(u'Creating cache file to %s' % cache_file)
            pickle.dump(CACHE, cache_fd, -1)
            pickle.dump(cache_stamp, cache_fd, -1)#}}}


def similar_artist(s_artist):#{{{unicodeReady
    """
    Retrieve similar artists on last.fm server and from user DB if available.
    <s_artist> is MPD current artist.
    similar_artist function returns a list.
    """
    as_artist_names = []
    # Look for AS ressources
    if s_artist in CACHE:
        as_artist_names = CACHE[s_artist]
        SIMA_LOGGER.debug(u'"%s" already in cache, not requesting last.fm.' %
                          unicode(s_artist,'utf-8'))
    else:
        try:
            as_artists = AudioScrobblerQuery(artist=s_artist)
        except AudioScrobblerConnectionError, as_error:
            SIMA_LOGGER.warning(u'Audioscrobbler server replies: %s…' %
                                as_error)
            return False

        for artist in as_artists.similar():
            as_artist_names.append(str(artist.name))
        CACHE[s_artist] = as_artist_names

    # Look for user ressources
    #
    if s_artist in USR_DB:
        as_artist_names = USR_DB[s_artist] + as_artist_names

    return as_artist_names#}}}

def top_tracks(s_artist):#{{{unicodeReady
    """
    Retrieve top tracks, ie. most popular song, from an artist.
    top_tracks function returns a list
    """
    top_tracks = []
    try:
        req = AudioScrobblerQuery(artist='s_artist')
    except AudioScrobblerConnectionError, as_error:
        SIMA_LOGGER.warning(u'Audioscrobbler server replies: %s…' %
                            as_error)
        return False
    top_tracks = [str(song.name) for song in req.toptracks()]#}}}

def sig_handler(signum, frame):#{{{unicodeReady
    """Catch signal other than KeyboardInterrupt"""
    raise KeyboardInterrupt(u"Caught a %d' SIG TERM signal" % signum)#}}}


def try_reconnect(mpd_con):#{{{unicodeReady
    mpd_con.disconnect()
    time.sleep(mpdConnector.WAIT_RECONNECT)
    if not mpd_main_connection(mpd_con):
        SIMA_LOGGER.warning(u'Remains not connected… waiting…')
    while(not mpd_con.is_connected()):
        time.sleep(mpdConnector.WAIT_RECONNECT)
        mpd_main_connection(mpd_con)#}}}


def mpd_main_connection(mpd_con):#{{{unicodeReady
    """
    Connect to mpd server.
    And controls available commands.
    """
    needed_cmds = ["status", "add", "find", "search", "currentsong", "ping"]
    if not mpd_con.connect():
        #sys.exit(2)
        return False
    elif mpd_con.is_connected():
        SIMA_LOGGER.info(u'Connected to MPD server (%s), proceeding…' %
                CONFIG.get('MPD', 'host'))

    try:
        CONFIG.getboolean('MPD', 'password')
        SIMA_LOGGER.info(u'No password set, proceeding without '+
                         u'authentification…')
    except ValueError:
        # ValueError if password not a boolean, hence an actual password.
        # pretty ugly? TODO: should I correct this?
        try:
            mpd_con.auth()
            SIMA_LOGGER.debug(u'Auth pass, proceeding…')
        except mpdConnector.MPDPasswordError, err:
            SIMA_LOGGER.error(u'Auth failed with "%s", wrong password?' % err)
            sys.exit(2)
    available_cmd = mpd_con.cmd()
    for nddcmd in needed_cmds:
        if nddcmd not in available_cmd:
            SIMA_LOGGER.error(u'command %s not available, '
                              u'control permissions…' % nddcmd)
            sys.exit(2)
    return True#}}}
#}}}

# Script starts here
###{{{unicodeReady
if __name__ == "__main__":
    #BOOT SEQUENCE {{{
    # Logging facility, default log level is INFO
    SIMA_LOGGER = sima_logger('debug')   #DEV# line used for debugging ;)
    #SIMA_LOGGER = sima_logger()
    SIMA_LOGGER.info(u'Starting MPD_sima version %s (revision %s)' % 
                     (__version__,__revison__))
    StartOpt(SIMA_LOGGER, (__version__, __revison__, __date__))

    # Configuration manager Object
    CONF_MANAGER = ConfMan(SIMA_LOGGER)
    CONFIG = CONF_MANAGER.control_conf()

    # Load Cache
    CACHE = {}
    if CONFIG.getboolean('sima', 'cache'):
        SIMA_LOGGER.info(u'Loading cache from:  %s' %
                         CONF_MANAGER.similar_cache)
        CACHE, CACHE_STAMP = load_cache(CONF_MANAGER.similar_cache)
        SIMA_LOGGER.info(u'%d artists in cache.' % len(CACHE))

    # Load user cache
    USR_DB = {}
    # USR_DB is a dict object, same format as CACHE
    #
    # USR_DB = {'artist O': ['similar 00', 'similar 01'],
    #           'artist 1': ['similar 01', 'similar 02']
    #           }
    # similar artist name 0 = None
    # similar artist name 1 = None
    # …
    # then USR_DB.items('artist') returns:
    # [ ('similar artist name 0',''), ('similar artist name 0','')]
    #
    if CONFIG.getboolean('sima', 'user_db'):
        SIMA_LOGGER.info(u'Loading user DB from:  %s' % CONF_MANAGER.user_cache)
        try:
            USR_DB = load_usr_cache(CONF_MANAGER.user_cache)
            #print [ art[0] for art in USR_DB.items('essai') ]
            print USR_DB
            SIMA_LOGGER.info(u'%d artists in user DB' % len(USR_DB))
        except CacheError, cache_err:
            SIMA_LOGGER.warning(cache_err)

    # Logging settings
    # Define the logger following user conf
    #  default log level is INFO.
    SIMA_LOGGER.setLevel(LEVELS.get(CONFIG.get('log', 'verbosity')))

    # MPD connection
    SIMA_LOGGER.debug(u'add %i tracks when playlist lenght <= %i track(s).' %
                      (CONFIG.getint('sima', 'track_to_add'),
                       CONFIG.getint('sima', 'queue_length')))
    SIMA_LOGGER.debug(u'About to connect to %s:%d, password "%s"' %
            (CONFIG.get('MPD', 'host'), CONFIG.getint('MPD', 'port'),
            unicode(CONFIG.get('MPD', 'password'),'utf-8')))

    # Create MPD object
    MPD_CON = mpdConnector.MpdConnector(CONFIG.get('MPD', 'host'), \
            CONFIG.getint('MPD', 'port'), CONFIG.get('MPD', 'password'))
    # Connect and auth if needed
    if not mpd_main_connection(MPD_CON):
        SIMA_LOGGER.error(u'Unable to connect to MPD on %s, exiting…' %
                CONFIG.get('MPD', 'host'))
        sys.exit(2)

    SIMA = SimaClass(CONFIG, MPD_CON)
    if CONFIG.getboolean('sima', 'history'):
        SIMA.inithist(CONF_MANAGER.history_file)
    else:
        SIMA_LOGGER.info(u'Do not load history, will not save session.')

    # In order to catch "kill 15" as KeyboardInterrupt when run on background
    signal.signal(signal.SIGTERM, sig_handler)
    #}}}

    while(1):#MAIN LOOP {{{
        try:
            if not MPD_CON.is_connected():
                SIMA_LOGGER.warning(u'Connexion with the MPD server lost.')
                SIMA_LOGGER.warning(u'Waiting %is before trying to reconnect.'
                        % mpdConnector.WAIT_RECONNECT)
                MPD_CON.disconnect()
                time.sleep(mpdConnector.WAIT_RECONNECT)
                if not mpd_main_connection(MPD_CON):
                    SIMA_LOGGER.warning(u'Remains not connected… waiting…')
                while(not MPD_CON.is_connected()):
                    time.sleep(mpdConnector.WAIT_RECONNECT)
                    mpd_main_connection(MPD_CON)
            MPD_PLAY_STATE = MPD_CON.status()['state']
            if MPD_CON.is_connected() and MPD_PLAY_STATE == 'play':
                SIMA.sima_lp()
                time.sleep(CONFIG.getint('sima', 'main_loop_time'))
                #SIMA_LOGGER.debug('###################################' )
            elif MPD_CON.is_connected() and \
                    MPD_PLAY_STATE == 'stop' or \
                    MPD_PLAY_STATE == 'pause':
                WAIT_TO_RESUME = 10
                SIMA_LOGGER.info(u'MPD is not playing (%s)…' %
                                 MPD_PLAY_STATE)
                SIMA_LOGGER.info(u'Waiting %is to check again…' %
                                 WAIT_TO_RESUME)
                while MPD_PLAY_STATE in ('pause', 'stop') and\
                      MPD_CON.is_connected():
                    MPD_PLAY_STATE = MPD_CON.status()['state']
                    time.sleep(int(WAIT_TO_RESUME))
                if MPD_CON.is_connected():
                    SIMA_LOGGER.info(u'MPD is now playing, proceeding…')
        except ConnectionError:
            SIMA_LOGGER.error(u'Connection error with MPD server.')
        except KeyboardInterrupt, int_err:
            if CONFIG.getboolean('sima', 'history'):
                SIMA.writehist(CONF_MANAGER.history_file)
            if CONFIG.getboolean('sima', 'cache') and CACHE:
                writecache(CONF_MANAGER.similar_cache)
                if CACHE_STAMP:
                    DELTA = date.today() - CACHE_STAMP['DATE']
                    SIMA_LOGGER.info(u'Cache is %d days old.' % DELTA.days)
            SIMA_LOGGER.warning(u'Exiting… %s' % int_err)
            break#}}}

    # Disconnect from MPD server
    try:
        MPD_CON.disconnect()
        SIMA_LOGGER.info(u'Disconnected from MPD server…')
    finally:
        sys.exit(0)
#}}}

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab
